/**
 * Created on Jun 11, 2006
 * 
 * @author Brian Mulanda
 *
 */
package edu.ksu.cis.ndpc;

import edu.ksu.cis.viewer.EmptyTreeException;
import edu.ksu.cis.viewer.TreeInterface;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Stack;

public class ParseTree implements TreeInterface, Cloneable {

	private int size;
	private Symbol root;
	private LinkedList children;
	private boolean hasChildren;
	private ParseTree parent;

	public ParseTree(Symbol root) {
		this.parent = null;
		this.root = root;
		this.hasChildren = false;
		this.size = 1;
	}

	public ParseTree(Symbol root, ParseTree child) {
		this.root = root;
		if (root != null && child != null) {
			add(child);
		}
	}

	public ParseTree(Symbol root, ParseTree leftChild, ParseTree rightChild) {
		this.root = root;
		if (root != null || leftChild != null || rightChild != null) {
			add(leftChild);
			add(rightChild);
		}
	}

	public void addChild(ParseTree child) {
		addChild(child, false);
	}

	public boolean addChild(ParseTree child, boolean first) {
		if (child == null) {
			return false;
		}

		if (first) {
			addFirst(child);
		} else {
			add(child);
		}
		return true;
	}

	public int size() {
		return size;
	}

	/* TreeInterface methods: Begin */
	public Object getRoot() throws EmptyTreeException {
		return root;
	}

	public TreeInterface[] getChildren() throws EmptyTreeException {
		TreeInterface[] childArray = null;
		if (hasChildren) {
			childArray = new TreeInterface[children.size()];
			for (int i = 0; i < children.size(); ++i) {
				childArray[i] = (TreeInterface) children.get(i);
			}
		}
		return childArray;
	}

	public boolean isEmpty() {
		return false;
	}

	/* TreeInterface methods: End */

	public Symbol getRootNode() {
		return root;
	}

	public int getSubtreeCount() {
		return hasChildren ? children.size() : 0;
	}

	public ParseTree[] getSubtrees() {
		ParseTree[] subtreeArray = null;
		if (hasChildren) {
			subtreeArray = new ParseTree[children.size()];
			for (int i = 0; i < children.size(); ++i) {
				subtreeArray[i] = (ParseTree) children.get(i);
			}
		}
		return subtreeArray;
	}

	public ParseTree getLeftSubtree() {
		ParseTree leftSubtree = null;
		if (hasChildren && children.size() == 2) {
			leftSubtree = (ParseTree) children.getFirst();
		}
		return leftSubtree;
	}

	public ParseTree getRightSubtree() {
		ParseTree rightSubtree = null;
		if (hasChildren && children.size() == 2) {
			rightSubtree = (ParseTree) children.getLast();
		}
		return rightSubtree;
	}

	public ParseTree getSubtree() {
		ParseTree subtree = null;
		if (hasChildren && children.size() == 1) {
			subtree = (ParseTree) children.getFirst();
		}
		return subtree;
	}

	public String toString() {
		return root.toString();
	}

	public String allToString() {
		StringBuffer s = new StringBuffer(root.toString());
		if (hasChildren && children.size() > 0) {
			Iterator it = children.iterator();
			s = s.append('(').append(((ParseTree) it.next()).allToString());
			while (it.hasNext()) {
				s.append(',').append(((ParseTree) it.next()).allToString());
			}
			s.append(')');
		}
		return s.toString();
	}

	public ParseTree getParent() {
		return parent;
	}

	public boolean replaceWith(ParseTree tree) {
		boolean replaced = false;
		if (tree != null) {
			this.root = (Symbol) tree.root.clone();
			if (tree.hasChildren) {
				if (this.hasChildren) {
					this.children.clear();
				} else {
					this.children = new LinkedList();
				}
				for (int i = 0; i < tree.children.size(); ++i) {
					ParseTree child = (ParseTree) ((ParseTree) tree.children
							.get(i)).clone();
					child.parent = this;
					this.children.add(child.clone());
				}
			} else {
				this.children = null;
			}
			this.hasChildren = (this.children != null);
			if (this.parent != null) {
				this.parent.size += (tree.size - this.size);
			}
			this.size = tree.size;
			replaced = true;
		}
		return replaced;
	}

	public int getPosition() {
		int position = 0;

		if (parent != null) {
			if (parent.hasChildren) {
				for (int i = 0; i < parent.children.size(); ++i) {
					if (parent.children.get(i) == this) {
						position = i;
					}
				}
			}
		}

		return position;
	}

	public int[] getPath() {
		int[] pathArr = null;
		Stack path = new Stack();
		if (parent != null) {
			ParseTree thisTree = this;

			while (thisTree.parent != null) {
				path.push(new Integer(thisTree.getPosition()));
				thisTree = thisTree.parent;
			}
		}

		pathArr = new int[path.size()];
		int i = 0;
		while (!path.isEmpty()) {
			pathArr[i++] = ((Integer) path.pop()).intValue();
		}
		return pathArr;
	}

	public ParseTree getSubtree(int position) {
		ParseTree subtree = null;

		if (hasChildren && children.size() > position && position >= 0) {
			subtree = (ParseTree) children.get(position);
		}
		return subtree;
	}

	public Object clone() {
		ParseTree newTree = new ParseTree((Symbol) this.root.clone());

		newTree.hasChildren = hasChildren;
		newTree.size = size;
		if (hasChildren) {
			newTree.children = new LinkedList();
			for (int i = 0; i < children.size(); ++i) {
				ParseTree child = (ParseTree) children.get(i);
				newTree.children.add(child.clone());
			}
		}

		return newTree;
	}

	/* node must not be null */
	public boolean containsNode(ParseTree node) {
		boolean hasNode = false;

		if (this.equals(node)) {
			hasNode = true;
		} else {
			if (hasChildren) {
				for (int i = 0; !hasNode && i < children.size(); ++i) {
					ParseTree childNode = (ParseTree) children.get(i);
					hasNode = childNode.containsNode(node);
				}
			}
		}
		return hasNode;
	}

	/* leaf must not be null */
	public boolean containsLeaf(ParseTree leaf) {
		boolean hasLeaf = false;

		if (isLeaf() && this.equals(leaf)) {
			hasLeaf = true;
		} else {
			if (hasChildren) {
				for (int i = 0; !hasLeaf && i < children.size(); ++i) {
					ParseTree node = (ParseTree) children.get(i);
					hasLeaf = node.containsLeaf(leaf);
				}
			}
		}
		return hasLeaf;
	}

	public boolean isLeaf() {
		return !hasChildren;
	}

	public boolean equals(Object obj) {
		boolean equal = true;
		try {
			if (obj instanceof ParseTree) {
				ParseTree tree = (ParseTree) obj;
				if (!tree.root.equals(this.root)) {
					equal = false; // Root Nodes not equal
				} else if (tree.children != null && this.children != null) {
					if (tree.children.size() != this.children.size()) {
						equal = false; // Number of children not equal
					} else {
						// Root Nodes equal, No. of children equal
						for (int i = 0; i < this.children.size(); ++i) {
							ParseTree child1 = (ParseTree) this.children.get(i);
							ParseTree child2 = (ParseTree) tree.children.get(i);
							if (!child1.equals(child2)) {
								equal = false;
								i = this.children.size();
							}
						}
					}
				}
			} else {
				equal = false;
			}
		} catch (Exception e) {
			equal = false;
		}
		return equal;
	}

	private void add(ParseTree child) {
		if (!hasChildren) {
			children = new LinkedList();
			hasChildren = true;
		}
		children.add(child);
		child.parent = this;
		size += child.size;
	}

	private void addFirst(ParseTree child) {
		if (!hasChildren) {
			children = new LinkedList();
			hasChildren = true;
		}
		children.addFirst(child);
		child.parent = this;
		size += child.size;
	}

	/*
	 * public int recomputeSize() { size = 1; if (hasChildren) { for (int i = 0;
	 * i < children.size(); ++i) { size +=
	 * ((ParseTree)children.get(i)).recomputeSize(); //size +=
	 * ((ParseTree)children.get(i)).size(); } } return size; }
	 */
}