package edu.ksu.cis.ndpc;

/**
 * @author Brian Mulanda
 *
 */
import java.util.Stack;
import java.util.ArrayList;

public class PredLParser implements LangParser {
	private PredLSymbols pSymbols;
	private boolean[][] syntaxTable;
	private ArrayList variables;

	public PredLParser() {
		initializeSyntaxTable();
		pSymbols = new PredLSymbols();
	}

	public LangSymbols getSymbols() {
		return pSymbols;
	}

	public void parse(ProofFormula pf) throws ParseException {
		String formula = pf.getFormulaString();
		variables = new ArrayList();
		ParseTree tree = null;
		if (formula != null && formula.trim().length() > 0) {
			formula = formula.trim();
			if (pSymbols.identify(formula) == LangSymbols.S_BOTTOM) {
				tree = new ParseTree(new Symbol(formula, LangSymbols.S_BOTTOM));
			} else {
				try {
					tree = parseFormula(new SymbolTokenizer(formula, pSymbols));
				} catch (ParseException prse) {
					throw new ParseException("Error in formula: " + formula
							+ "\n" + prse.getMessage());
				}
			}
		}

		if (tree != null) {
			pf.tree = tree;
			if (variables.size() > 0) {
				pf.variables = variables;
			}
		}
	}

	private ParseTree parseFormula(SymbolTokenizer tokenizer)
			throws ParseException {
		return parseFormula(tokenizer, false, false);
	}

	private ParseTree parseFormula(SymbolTokenizer tokenizer,
			boolean doneOnClose, boolean oneFormula) throws ParseException {

		Stack opStack = new Stack();
		Stack treeStack = new Stack();

		int pCount = 0;

		boolean parenth = true;
		boolean done = false;
		boolean expectEq = false;
		ParseTree tree = null;

		Symbol inputSymbol = null;
		int inputType = LangSymbols.S_NULL;
		int recentType = LangSymbols.S_NULL;
		int recentOpType = LangSymbols.S_NULL;

		while (!done) {
			if (!tokenizer.nextSymbol()) {
				done = true;
				inputType = LangSymbols.S_NULL;
				if (!syntaxTable[recentType][inputType]) {
					throw new ParseException(" Incomplete formula");
				}
			} else {
				inputSymbol = tokenizer.getSymbol();
				inputType = inputSymbol.getType();
				if (inputType == LangSymbols.S_INV) {
					throw new ParseException(" Invalid symbol " + inputSymbol);
				} else {
					if (syntaxTable[recentType][inputType]) {
						if (inputType == LangSymbols.S_PRED) {
							treeStack.push(parsePredicate(tokenizer,
									inputSymbol));
						} else if (inputType == LangSymbols.S_OPAR) {
							pCount++;
							opStack.push(inputSymbol);
							parenth = true;
						} else if (inputType == LangSymbols.S_CPAR) {
							--pCount;
							if (pCount < 0) {
								throw new ParseException(" Missing a (");
							} else if (pCount == 0) {
								parenth = false;
								if (doneOnClose) {
									done = true;
								}
							}
							constructTree(opStack, treeStack, inputType);
							if (!opStack.isEmpty()) {
								Symbol symbol = (Symbol) opStack.peek();
								if (pSymbols.isOperator(symbol)) {
									recentOpType = symbol.getType();
								} else {
									recentOpType = LangSymbols.S_NULL;
								}
							} else {
								recentOpType = LangSymbols.S_NULL;
							}
						} else {
							if (PredLSymbols.isTerm(inputType)) {
								if (inputType == LangSymbols.S_FUNC) {
									treeStack.push(parseFunction(tokenizer,
											inputSymbol));
								} else {
									treeStack.push(new ParseTree(inputSymbol));
								}
								if (expectEq) {
									ParseTree tempTree = new ParseTree(
											(Symbol) opStack.pop());
									tempTree.addChild((ParseTree) treeStack
											.pop(), true);
									tempTree.addChild((ParseTree) treeStack
											.pop(), true);
									treeStack.push(tempTree);
									inputType = LangSymbols.S_PRED;
								}
								expectEq = !expectEq;
							} else if (inputType == LangSymbols.S_EXISTS
									|| inputType == LangSymbols.S_FORALL) {
								treeStack.push(parseQuantifier(tokenizer,
										inputSymbol));
								inputType = LangSymbols.S_PRED;
								if (oneFormula && pCount == 0) {
									done = true;
								}
							} else if (pSymbols.isOperator(inputType)) {
								if (!parenth) {
									if (checkOperatorSyntax(recentOpType,
											inputType)) {
										if (buildOp(recentOpType, inputType)) {
											constructTree(opStack, treeStack,
													inputType);
										}
									} else {
										throw new ParseException(
												" Ambiguous use of operators");
									}
								}
								opStack.push(inputSymbol);
								recentOpType = inputType;
								parenth = false;
							} else {
								throw new ParseException(" Invalid symbol "
										+ inputSymbol.getValue());
							}
						}
						if (inputType == LangSymbols.S_PRED && oneFormula
								&& pCount == 0) {
							done = true;
						} else {
							recentType = inputType;
						}
					} else {
						throw new ParseException(" Unexpected symbol "
								+ inputSymbol);
					}
				}
			}
		}

		if (pCount > 0) {
			throw new ParseException(" Missing a )");
		} else if (expectEq) {
			throw new ParseException(" Expecting =");
		}

		if (!opStack.isEmpty()) {
			constructTree(opStack, treeStack, LangSymbols.S_NULL);
		}
		if (!treeStack.isEmpty()) {
			tree = (ParseTree) treeStack.pop();
		}

		// System.out.println("End Parse ProofFormula: pCount="+pCount);
		return tree;
	}

	private ParseTree parseQuantifier(SymbolTokenizer tokenizer,
			Symbol quantifier) throws ParseException {
		ParseTree tree = null;
		if (tokenizer.nextSymbol()) {
			Symbol inputSymbol = tokenizer.getSymbol();
			int type = inputSymbol.getType();

			if (type == LangSymbols.S_VAR) {
				variables.add(inputSymbol);
				tree = new ParseTree(quantifier.concat(inputSymbol));
				ParseTree tempTree = parseFormula(tokenizer, true, true);
				tree.addChild(tempTree);
			} else if (type == LangSymbols.S_CONST) {
				throw new ParseException(inputSymbol
						+ " is a constant; constants cannot be quantified by "
						+ quantifier + ".");
			} else {
				if (type != LangSymbols.S_INV) {
					throw new ParseException(inputSymbol
							+ " cannot be used here '" + quantifier
							+ inputSymbol + "'; use a variable instead.");
				} else {
					throw new ParseException(inputSymbol
							+ " is not a valid symbol.");
				}
			}
		}
		return tree;
	}

	private ParseTree parseFunction(SymbolTokenizer tokenizer, Symbol function)
			throws ParseException {
		ParseTree tree = new ParseTree(function);
		parseArguments(tokenizer, tree);

		int givenArity = tree.getSubtreeCount();
		int specArity = pSymbols.getArity(function);
		if (specArity >= 1) {
			if (specArity != givenArity) {
				throw new ParseException("Function '" + function + "' expects "
						+ specArity + " argument(s); you have supplied "
						+ givenArity + " argument(s)");
			}
		} else {
			pSymbols.addSymbol(function, givenArity);
		}
		return tree;
	}

	private ParseTree parsePredicate(SymbolTokenizer tokenizer, Symbol predicate)
			throws ParseException {
		ParseTree tree = new ParseTree(predicate);
		if (tokenizer.nextSymbol(true)) {
			if (tokenizer.getSymbol(true).getType() == LangSymbols.S_OPAR) {
				parseArguments(tokenizer, tree);
			}
		}

		int givenArity = tree.getSubtreeCount();
		int specArity = pSymbols.getArity(predicate);
		if (specArity >= 0) {
			if (specArity != givenArity) {
				throw new ParseException("Predicate '" + predicate
						+ "' expects " + specArity
						+ " argument(s); you have supplied " + givenArity
						+ " argument(s)");
			}
		} else {
			pSymbols.addSymbol(predicate, givenArity);
		}
		return tree;
	}

	private void parseArguments(SymbolTokenizer tokenizer, ParseTree parent)
			throws ParseException {

		boolean done = false;
		Symbol inputSymbol = null;
		int recentType = LangSymbols.S_NULL;
		int inputType = LangSymbols.S_NULL;
		int pCount = 0;

		while (!done) {
			if (!tokenizer.nextSymbol()) {
				if (recentType != LangSymbols.S_NULL) {
					if (pCount > 0) {
						throw new ParseException("Missing )");
						/*
						 * if ( recentType == LangSymbols.S_CONST || recentType
						 * == LangSymbols.S_VAR ) {
						 * 
						 * throw new ParseException("Missing )"); } else { throw
						 * new ParseException("Unexpected end of input"); }
						 */
					} else {
						throw new ParseException("Unexpected end of input");
					}
				} else {
					throw new ParseException("Unexpected end of input");
				}
			} else {
				inputSymbol = tokenizer.getSymbol();
				inputType = inputSymbol.getType();
				if (inputType == LangSymbols.S_INV) {
					throw new ParseException("Invalid symbol " + inputSymbol);
				} else {
					if (checkArgumentSyntax(recentType, inputType)) {
						switch (inputType) {
						case LangSymbols.S_CONST:
							parent.addChild(new ParseTree(inputSymbol));
							pSymbols.addSymbol(inputSymbol);
							break;
						case LangSymbols.S_VAR:
							variables.add(inputSymbol);
							parent.addChild(new ParseTree(inputSymbol));
							pSymbols.addSymbol(inputSymbol);
							break;
						case LangSymbols.S_FUNC:
							parent.addChild(parseFunction(tokenizer,
									inputSymbol));
							break;
						case LangSymbols.S_OPAR:
							pCount++;
							break;
						case LangSymbols.S_CPAR:
							pCount--;
							if (pCount == 0) {
								done = true;
							}
							break;
						case LangSymbols.S_COMMA:
							break;
						default:

						}
						recentType = inputType;
					} else {
						throw new ParseException(" Unexpected symbol "
								+ inputSymbol);
					}
				}
			}
		}
	}

	private boolean checkArgumentSyntax(int recentType, int inputType) {
		boolean syntaxOk = false;

		if (inputType == LangSymbols.S_NULL) {
			if (recentType == LangSymbols.S_CPAR) {
				syntaxOk = true;
			}
		} else {
			if (recentType == LangSymbols.S_NULL
					|| recentType == LangSymbols.S_COMMA) {
				if (PredLSymbols.isTerm(inputType)
						|| inputType == LangSymbols.S_OPAR) {
					syntaxOk = true;

				}
			} else if (pSymbols.isConstant(recentType)
					|| pSymbols.isVariable(recentType)) {
				if (inputType == LangSymbols.S_CPAR
						|| inputType == LangSymbols.S_COMMA) {
					syntaxOk = true;
				}
			} else if (pSymbols.isFunction(recentType)) {
				if (inputType == LangSymbols.S_CPAR
						|| inputType == LangSymbols.S_COMMA) {
					syntaxOk = true;
				}
			} else if (recentType == LangSymbols.S_OPAR) {
				if (inputType == LangSymbols.S_OPAR
						|| PredLSymbols.isTerm(inputType)) {
					syntaxOk = true;
				}
			} else if (recentType == LangSymbols.S_CPAR) {
				if (inputType == LangSymbols.S_CPAR
						|| inputType == LangSymbols.S_COMMA) {
					syntaxOk = true;
				}
			}
		}
		return syntaxOk;
	}

	private boolean buildOp(int stackOpType, int inputOpType)
			throws ParseException {

		boolean build = false;
		if (stackOpType != LangSymbols.S_NULL) {

			if ((stackOpType == LangSymbols.S_FORALL && inputOpType == LangSymbols.S_EXISTS)
					|| (stackOpType == LangSymbols.S_EXISTS && inputOpType == LangSymbols.S_FORALL)) {

				return build;
			} else if (pSymbols.getRelativePrecedence(stackOpType, inputOpType) == LangSymbols.PRECE_EQUAL) {
				throw new ParseException(" Ambiguous operator precedence for "
						+ stackOpType + " and " + inputOpType);
			} else if (pSymbols.getRelativePrecedence(stackOpType, inputOpType) == LangSymbols.PRECE_HIGHER) {
				build = true;
			} // precedence == EQUAL, HIGHER
		}
		return build;
	}

	private boolean checkOperatorSyntax(int recentOpType, int inputOpType) {
		boolean syntaxOk = false;

		if (recentOpType == LangSymbols.S_NULL) {
			syntaxOk = true;
		} else {
			if (pSymbols.isOperator(inputOpType)
					&& pSymbols.isOperator(recentOpType)) {
				// Both operators
				if (!(recentOpType == LangSymbols.S_CONJ && inputOpType == LangSymbols.S_DISJ)
						&& !(recentOpType == LangSymbols.S_DISJ && inputOpType == LangSymbols.S_CONJ)) {
					syntaxOk = true; // Ambiguous operator precedence v and ^
				}
			}
		}
		return syntaxOk;
	}

	private void constructTree(Stack opStack, Stack treeStack, int inputOpType)
			throws ParseException {

		try {
			boolean done = false;
			while (!done) {
				if (opStack.isEmpty()) {
					done = true;
				} else {
					Symbol opSymbol = (Symbol) opStack.peek();
					if (inputOpType == LangSymbols.S_CPAR) {
						if (opSymbol.getType() == LangSymbols.S_OPAR) {
							opStack.pop();
							done = true;
						}
					} else if (pSymbols.isOperator(inputOpType)) { // Is
																	// Operator
						if (pSymbols.isOperator(opSymbol)) {
							if (!buildOp(opSymbol.getType(), inputOpType)) {
								done = true;
							}
						} else {
							done = true;
						}
					}

					if (!done) {
						opStack.pop();
						ParseTree tree = null;
						if (opSymbol.getType() == LangSymbols.S_FORALL
								|| opSymbol.getType() == LangSymbols.S_EXISTS) {
							ParseTree child = (ParseTree) treeStack.pop();
							tree = (ParseTree) treeStack.pop();
							tree.addChild(child);
						} else if (opSymbol.getType() == LangSymbols.S_NEG) {
							tree = new ParseTree(opSymbol);
							tree.addChild((ParseTree) treeStack.pop());
						} else {
							tree = new ParseTree(opSymbol);
							tree.addChild((ParseTree) treeStack.pop(), true);
							tree.addChild((ParseTree) treeStack.pop(), true);
						}
						treeStack.push(tree);
					}
				}
			}
		} catch (ParseException pse) {
			throw new ParseException(pse.getMessage());
		} catch (Exception e) {
			throw new ParseException(e.getMessage());
			// e.printStackTrace();
		}
	}

	private void initializeSyntaxTable() {
		syntaxTable = new boolean[][] {

				/*
				 * Recent Current Input / Input / P -> <-> ^ v ~ ( ) EOF _|_ c f
				 * x A E = / P
				 */{ false, true, true, true, true, false, false, true, true,
						false, false, false, false, false, false, false },
				/* -> */{ true, false, false, false, false, true, true, false,
						false, false, true, true, true, true, true, false },
				/* <-> */{ true, false, false, false, false, true, true, false,
						false, false, true, true, true, true, true, false },
				/* ^ */{ true, false, false, false, false, true, true, false,
						false, false, true, true, true, true, true, false },
				/* v */{ true, false, false, false, false, true, true, false,
						false, false, true, true, true, true, true, false },
				/* ~ */{ true, false, false, false, false, true, true, false,
						false, false, true, true, true, true, true, false },
				/* ( */{ true, false, false, false, false, true, true, false,
						false, false, true, true, true, true, true, false },
				/* ) */{ true, true, true, true, true, false, false, true,
						true, false, false, false, false, false, false, true },
				/* EOF */{ true, false, false, false, false, true, true, false,
						false, true, true, true, true, true, true, false },
				/* _|_ */{ false, false, false, false, false, false, false,
						false, true, false, false, false, false, false, false,
						false },
				/* c */{ false, false, false, false, false, false, false, true,
						true, false, false, false, false, false, false, true },
				/* f */{ false, false, false, false, false, false, false, true,
						true, false, false, false, false, false, false, true },
				/* x */{ false, false, false, false, false, false, false, true,
						true, false, false, false, false, false, false, true },
				/* A */{ true, false, false, false, false, true, true, false,
						false, false, true, true, true, true, true, false },
				/* E */{ true, false, false, false, false, true, true, false,
						false, false, true, true, true, true, true, false },
				/* = */{ false, false, false, false, false, false, true, false,
						false, false, true, true, true, false, false, false }

		};
	}
}