package edu.ksu.cis.ndpc;

import java.util.LinkedList;
import java.util.ArrayList;

import edu.ksu.cis.ndpc.rules.RuleException;

/**
 * @author Brian Mulanda
 * 
 */
public class Proof {
	private LinkedList statements;
	private ProofTree tree;
	private ProofTree currentTree;
	private int boxId;
	private int nodeId;
	private int count;
	private int lineno;
	private ArrayList variables;
	private ArrayList freshVariables;

	public Proof() {
		statements = new LinkedList();
		tree = new ProofTree(new ProofTreeNode(new ProofBox(0),
				Integer.MAX_VALUE));
		currentTree = tree;
		boxId = 1;
		nodeId = 1;
		variables = new ArrayList();
		freshVariables = new ArrayList();
	}

	public void setActualStatCount(int count) {
		this.count = count;
	}

	public int getActualStatCount() {
		return count;
	}

	public void setLastStatNo(int lineno) {
		this.lineno = lineno;
	}

	public int getLastStatNo() {
		return lineno;
	}

	public void addStatement(ProofStatement statement) {
		ProofTree statementTree = new ProofTree(new ProofTreeNode(statement,
				nodeId++));
		currentTree.addChild(statementTree);
		statements.add(statementTree);
	}

	public void addVariable(Symbol variable) {
		if (variable != null)
			variables.add(variable);
	}

	public void addFreshVariable(Symbol variable, int lineno) {
		if (variable != null) {
			freshVariables.add(variable);
			ProofTree statTree = getStatementTree(lineno);
			((ProofBox) statTree.getParent().getRootNode().getNode())
					.addFreshVariable(variable);
		}
	}

	public void openBox() {
		ProofBox box = new ProofBox(boxId++);
		ProofTree boxTree = new ProofTree(new ProofTreeNode(box, nodeId++));
		currentTree.addChild(boxTree);
		currentTree = boxTree;
	}

	public void closeBox() {
		((ProofTreeNode) currentTree.getRoot()).setNodeId(nodeId++);
		currentTree = currentTree.getParent();
	}

	public ProofTree getTree() {
		return tree;
	}

	public int getStatementCount() {
		return statements.size();
	}

	public ProofStatement getStatement(int lineno) {
		return getStatement(getStatementTree(lineno));
	}

	public BoxedStatements getBoxedStatements(Range range)
			throws ProofException {
		int begin = range.getBegin();
		int end = range.getEnd();

		BoxedStatements statements = new BoxedStatements(range);
		statements.setFirst(getStatement(begin));
		statements.setLast(getStatement(end));
		return statements;
	}

	public boolean canReferenceBox(int line, LineReference range)
			throws ProofException {
		boolean ok = false;
		int begin = range.getBegin();
		int end = range.getEnd();

		if (line > end) {
			ProofTree boxBegin = getContainingBox(begin);
			ProofTree boxEnd = getContainingBox(end);

			if (boxBegin != boxEnd) {
				throw new ProofException("Statements " + begin + " and " + end
						+ " must be " + "contained within the same box.");
			}

			int statNodeId = getStatementTree(line).getRootNode().getNodeId();
			int boxNodeId = boxBegin.getParent().getRootNode().getNodeId();

			if (statNodeId < boxNodeId) {
				ok = true;
			}
		}
		return ok;
	}

	public boolean statIsInBox(int line) {
		return (getContainingBox(line) != null);
	}

	public boolean statIsFirstInBox(int line) {
		boolean first = false;
		ProofTree boxTree = getContainingBox(line);
		if (boxTree != null) {
			ProofTree[] subtrees = boxTree.getSubtrees();
			for (int i = 0; i < subtrees.length; ++i) {
				Object node = subtrees[i].getRootNode().getNode();
				if (node instanceof ProofStatement) {
					if (node == getStatement(line)) {
						first = true;
					}
					i = subtrees.length;
				}
			}
		}
		return first;
	}

	private ProofTree getContainingBox(int line) {
		ProofTree boxTree = null;

		ProofTree tree = getStatementTree(line);
		if (tree != null) {
			tree = tree.getParent();
			if (tree.getParent() != null) {
				boxTree = tree;
			}
		}
		return boxTree;
	}

	public boolean canReferenceLine(int line, int refline) {
		boolean ok = false;
		ProofTree lineTree = null;
		ProofTree lineRefTree = null;

		if (line > refline) {
			lineTree = getStatementTree(line);
			lineRefTree = getStatementTree(refline);
			if (lineTree != null && lineRefTree != null) {
				if (lineRefTree.getParent().getRootNode().getNodeId() > lineTree
						.getRootNode().getNodeId()) {
					ok = true;
				}
			}
		}
		return ok;
	}

	public boolean rangeIsBox(LineReference range) {
		boolean ok = false;
		if (range.getBegin() > 0 && range.getEnd() > 0
				&& range.getBegin() <= range.getEnd()) {

			ProofTree beginTree = this.getContainingBox(range.getBegin());
			ProofTree endTree = this.getContainingBox(range.getEnd());
			if (beginTree != null && endTree != null
					&& beginTree.getParent() != null
					&& endTree.getParent() != null) {

				int begin = getStatement(beginTree.getFirstSubtree())
						.getLineNo();
				int end = getStatement(endTree.getLastSubtree()).getLineNo();
				ok = (range.getBegin() == begin && range.getEnd() == end);
			}
		}
		return ok;
	}

	public boolean isUsedVariable(Symbol variable) {
		return (freshVariables.contains(variable) || variables
				.contains(variable));
	}

	public boolean isFreshVariable(Symbol variable) {
		return freshVariables.contains(variable);
	}

	public boolean isInScope(Symbol variable, int lineno) {
		boolean inscope = false;

		ProofTree tree = getStatementTree(lineno).getParent();
		while (tree != null) {
			ProofBox box = (ProofBox) tree.getRootNode().getNode();
			if (box.introduces(variable)) {
				inscope = true;
				tree = null;
			} else {
				tree = tree.getParent();
			}
		}
		return inscope;
	}

	private ProofStatement getStatement(ProofTree statement) {
		if (statement != null) {
			return (ProofStatement) statement.getRootNode().getNode();
		} else {
			return null;
		}
	}

	private ProofTree getStatementTree(int lineno) {
		ProofTree statement = null;
		int index = lineno - 1;
		if (index >= 0 && index < statements.size()) {
			statement = (ProofTree) statements.get(index);
		}
		return statement;
	}
}