package edu.ksu.cis.ndpc;

/**
 * @author Brian Mulanda
 *
 */
import java.util.Stack;

public class PropLParser implements LangParser {
	private PropLSymbols symbols;
	private boolean[][] syntaxTable;

	public PropLParser() {
		initializeSyntaxTable();
	}

	public LangSymbols getSymbols() {
		return symbols;
	}

	public void parse(ProofFormula pf) throws ParseException {
		String formula = pf.formula;
		symbols = new PropLSymbols();
		ParseTree tree = null;
		if (formula != null && formula.trim().length() > 0) {
			formula = formula.trim();
			if (symbols.identify(formula) == LangSymbols.S_BOTTOM) {
				tree = new ParseTree(new Symbol(formula, LangSymbols.S_BOTTOM));
			} else {
				try {
					tree = parseFormula(new SymbolTokenizer(formula, symbols));
				} catch (ParseException prse) {
					throw new ParseException("Error in formula: " + formula
							+ "\n" + prse.getMessage());
				}
			}
		}

		if (tree != null) {
			pf.tree = tree;
		}
	}

	private ParseTree parseFormula(SymbolTokenizer tokenizer)
			throws ParseException {
		return parseFormula(tokenizer, false);
	}

	private ParseTree parseFormula(SymbolTokenizer tokenizer,
			boolean doneOnClose) throws ParseException {

		Stack opStack = new Stack();
		Stack treeStack = new Stack();

		int pCount = 0;

		boolean parenth = true;
		boolean done = false;
		ParseTree tree = null;

		Symbol inputSymbol = null;
		int inputType = LangSymbols.S_NULL;
		int recentType = LangSymbols.S_NULL;
		int recentOpType = LangSymbols.S_NULL;

		while (!done) {
			if (!tokenizer.nextSymbol()) {
				done = true;
				inputType = LangSymbols.S_NULL;
				if (!syntaxTable[recentType][inputType]) {
					throw new ParseException(" Incomplete formula");
				}
			} else {
				inputSymbol = tokenizer.getSymbol();
				inputType = inputSymbol.getType();
				if (inputType == LangSymbols.S_INV) {
					throw new ParseException(" Invalid symbol " + inputSymbol);
				} else {
					if (syntaxTable[recentType][inputType]) {
						if (inputType == LangSymbols.S_ATOM) {
							treeStack.push(new ParseTree(inputSymbol));
						} else if (inputType == LangSymbols.S_OPAR) {
							pCount++;
							opStack.push(inputSymbol);
							parenth = true;
						} else if (inputType == LangSymbols.S_CPAR) {
							--pCount;
							if (pCount < 0) {
								throw new ParseException("Missing a (");
							} else if (pCount == 0) {
								parenth = false;
								if (doneOnClose) {
									done = true;
								}
							}
							constructTree(opStack, treeStack, inputType);
							if (!opStack.isEmpty()) {
								Symbol symbol = (Symbol) opStack.peek();
								if (symbols.isOperator(symbol)) {
									recentOpType = symbol.getType();
								} else {
									recentOpType = LangSymbols.S_NULL;
								}
							} else {
								recentOpType = LangSymbols.S_NULL;
							}
						} else {
							if (symbols.isOperator(inputType)) {
								if (!parenth) {
									if (checkOperatorSyntax(recentOpType,
											inputType)) {
										if (buildOp(recentOpType, inputType)) {
											constructTree(opStack, treeStack,
													inputType);
										}
									} else {
										// throw new
										// ParseException(" parse:Failed operator syntax");
										throw new ParseException(
												" Ambiguous use of operators");
									}
								}

								opStack.push(inputSymbol);
								recentOpType = inputType;
								parenth = false;
							} else {
								throw new ParseException(" Invalid symbol "
										+ inputSymbol.getValue());
							}
						}
						recentType = inputType;
					} else {
						throw new ParseException(" Unexpected symbol "
								+ inputSymbol);
					}
				}
			}
		}

		if (pCount > 0) {
			throw new ParseException("Missing a )");
		}

		if (!opStack.isEmpty()) {
			constructTree(opStack, treeStack, LangSymbols.S_NULL);
		}

		if (!treeStack.isEmpty()) {
			tree = (ParseTree) treeStack.pop();
		}

		return tree;
	}

	private boolean buildOp(int stackOpType, int inputOpType)
			throws ParseException {

		boolean build = false;
		if (stackOpType != LangSymbols.S_NULL) {
			if (symbols.getRelativePrecedence(stackOpType, inputOpType) == LangSymbols.PRECE_EQUAL) {
				throw new ParseException("Ambiguous operator precedence for "
						+ stackOpType + " and " + inputOpType);
			} else if (symbols.getRelativePrecedence(stackOpType, inputOpType) == LangSymbols.PRECE_HIGHER) {
				build = true;
			} // precedence == EQUAL, HIGHER
		}
		return build;
	}

	private boolean checkOperatorSyntax(int recentOpType, int inputOpType) {
		boolean syntaxOk = false;

		if (recentOpType == LangSymbols.S_NULL) {
			syntaxOk = true;
		} else {
			if (symbols.isOperator(inputOpType)
					&& symbols.isOperator(recentOpType)) {
				// Both operators
				if (!(recentOpType == LangSymbols.S_CONJ && inputOpType == LangSymbols.S_DISJ)
						&& !(recentOpType == LangSymbols.S_DISJ && inputOpType == LangSymbols.S_CONJ)) {
					syntaxOk = true; // Non-Ambiguous operator precedence v and
										// ^
				}
			}
		}
		return syntaxOk;
	}

	private void constructTree(Stack opStack, Stack treeStack, int inputOpType)
			throws ParseException {

		try {
			boolean done = false;
			while (!done) {
				if (opStack.isEmpty()) {
					done = true;
				} else {
					Symbol opSymbol = (Symbol) opStack.peek();
					if (inputOpType == LangSymbols.S_CPAR) {
						if (opSymbol.getType() == LangSymbols.S_OPAR) {
							opStack.pop();
							done = true;
						}
					} else if (symbols.isOperator(inputOpType)) { // Is Operator
						if (symbols.isOperator(opSymbol)) {
							if (!buildOp(opSymbol.getType(), inputOpType)) {
								done = true;
							}
						} else {
							done = true;
						}
					}

					if (!done) {
						opStack.pop();
						ParseTree tree = null;
						if (opSymbol.getType() == LangSymbols.S_NEG) {
							tree = new ParseTree(opSymbol);
							tree.addChild((ParseTree) treeStack.pop());
						} else {
							tree = new ParseTree(opSymbol);
							tree.addChild((ParseTree) treeStack.pop(), true);
							tree.addChild((ParseTree) treeStack.pop(), true);
						}
						treeStack.push(tree);
					}
				}
			}
		} catch (ParseException pse) {
			throw new ParseException(pse.getMessage());
		} catch (Exception e) {
			throw new ParseException(e.getMessage());
			// e.printStackTrace();

		}
	}

	private void initializeSyntaxTable() {
		syntaxTable = new boolean[][] {

				/*
				 * Recent Current Input / Input / p -> <-> ^ v ~ ( ) EOF _|_ / p
				 */{ false, true, true, true, true, false, false, true, true,
						false },
				/* -> */{ true, false, false, false, false, true, true, false,
						false, false },
				/* <-> */{ true, false, false, false, false, true, true, false,
						false, false },
				/* ^ */{ true, false, false, false, false, true, true, false,
						false, false },
				/* v */{ true, false, false, false, false, true, true, false,
						false, false },
				/* ~ */{ true, false, false, false, false, true, true, false,
						false, false },
				/* ( */{ true, false, false, false, false, true, true, false,
						false, false },
				/* ) */{ false, true, true, true, true, false, false, true,
						true, false },
				/* EOF */{ true, false, false, false, false, true, true, false,
						false, false },
				/* _|_ */{ false, false, false, false, false, false, false,
						false, false, false }

		};
	}
}
