/*
 * IO.java       Jan 19, 2007
 *
 * Copyright (c) 2007, Brian Mulanda, all rights reserved.
 */

package edu.ksu.cis.ndpc.io;

import java.io.File;
import java.io.IOException;
import java.awt.Component;

/**
 * A class for performing general IO operations.
 * 
 * @author Brian Mulanda (<a
 *         href="mailto:bmulanda@ksu.edu">bmulanda@ksu.edu</a>)
 * 
 * @see InputFile
 * @see OutputFile
 * @see CustomFileChooser
 * 
 */

public class IO {
	public static String currentDirectory = ".";

	public static final int FILE_OPEN = 0;
	public static final int FILE_SAVE = 1;
	public static final int FILE_SAVEAS = 2;
	public static final int FILE_EXPORT = 3;
	public static final int FILE_EXPORT_IMAGE = 4;

	public static final int TYPE_UNKNOWN = -1;
	public static final int TYPE_OBJECT = 0;
	public static final int TYPE_TEXT = 1;

	public static File chooseFile(int type) {
		return chooseFile(type, null);
	}

	public static File chooseFile(int type, Component parent) {
		File file = null;
		int returnVal = CustomFileChooser.CANCEL_OPTION;
		CustomFileChooser fc = getFileChooser(type);
		if (type == FILE_OPEN) {
			returnVal = fc.showOpenDialog(parent);
		} else if (type == FILE_SAVE) {
			returnVal = fc.showSaveDialog(parent);
		} else if (type == FILE_SAVEAS) {
			returnVal = fc.showDialog(parent, "Save As");
		} else if (type == FILE_EXPORT) {
			returnVal = fc.showDialog(parent, "Save As");
		} else if (type == FILE_EXPORT_IMAGE) {
			returnVal = fc.showDialog(parent, "Save As");
		}
		if (returnVal == CustomFileChooser.APPROVE_OPTION) {
			file = fc.getSelectedFile();
			String fileExt = CustomFileFilter.getExtension(file);
			String selectExt = CustomFileFilter
					.getExtension((CustomFileFilter) fc.getFileFilter());
			if (fileExt == null || !fileExt.equalsIgnoreCase(selectExt)) {
				file = new File(file.getParentFile().getAbsolutePath()
						+ File.separator + file.getName() + "." + selectExt);
			}
			IO.currentDirectory = file.getParent();
		}
		fc = null;
		return file;
	}

	public static CustomFileChooser getFileChooser(int type) {
		CustomFileChooser fc = new CustomFileChooser(IO.currentDirectory);
		fc.setAcceptAllFileFilterUsed(false);
		if (type == FILE_OPEN) {
			fc.setDialogTitle("Open Proof");
			fc.setType(CustomFileChooser.TYPE_OPEN);
			fc
					.addChoosableFileFilter(new CustomFileFilter("txt",
							"Text Files"));
			fc.addChoosableFileFilter(new CustomFileFilter("ndp",
					"Natural Deduction Proofs"));
		} else if (type == FILE_SAVEAS || type == FILE_SAVE) {
			fc.addChoosableFileFilter(new CustomFileFilter("txt", "Text File"));
			fc.addChoosableFileFilter(new CustomFileFilter("ndp",
					"Natural Deduction Proof"));
			if (type == FILE_SAVE) {
				fc.setDialogTitle("Save Proof");
				fc.setType(CustomFileChooser.TYPE_SAVE);
			} else if (type == FILE_SAVEAS) {
				fc.setDialogTitle("Save Proof As");
				fc.setType(CustomFileChooser.TYPE_SAVEAS);
			}
		} else if (type == FILE_EXPORT) {
			fc.setDialogTitle("Export Proof");
			fc.setType(CustomFileChooser.TYPE_SAVEAS);
			fc
					.addChoosableFileFilter(new CustomFileFilter("tex",
							"LaTeX Files"));
		} else if (type == FILE_EXPORT_IMAGE) {
			fc.setDialogTitle("Export Proof as Image");
			fc.setType(CustomFileChooser.TYPE_SAVEAS);
			fc.addChoosableFileFilter(new CustomFileFilter("bmp", "BMP"));
			fc.addChoosableFileFilter(new CustomFileFilter("jpg", "JPEG"));
			fc.addChoosableFileFilter(new CustomFileFilter("png", "PNG "));
		}
		return fc;
	}

	public static InputFile openInputFile(String filename) {
		return openInputFile(new File(filename));
	}

	public static InputFile openInputFile(File f) {
		InputFile infile = null;
		if (f != null) {
			try {
				infile = new InputFile(f);
			} catch (Exception e) {
			}
		}
		return infile;
	}

	public static InputFile openInputFile(Component parent) {
		return openInputFile(chooseFile(FILE_OPEN, parent));
	}

	public static OutputFile openOutputFile(String filename) {
		return openOutputFile(new File(filename));
	}

	public static OutputFile openOutputFile(File f) {
		return openOutputFile(f, false);
	}

	public static OutputFile openOutputFile(File f, boolean append) {
		OutputFile outfile = null;
		if (f != null) {
			try {
				outfile = new OutputFile(f, append);
			} catch (Exception e) {
			}
		}
		return outfile;
	}

	public static OutputFile openOutputFile(Component parent) {
		return openOutputFile(false, parent);
	}

	public static OutputFile openOutputFile(boolean append, Component parent) {
		return openOutputFile(chooseFile(FILE_SAVEAS, parent), false);
	}

	public static OutputFile openImageOutputFile(Component parent) {
		return openOutputFile(chooseFile(FILE_EXPORT_IMAGE, parent), false);
	}

	public static void writeToFile(OutputFile file, Object data)
			throws IOException {
		int type = getType(file);
		if (type == TYPE_TEXT) {
			file.writeSting(data.toString());
		} else if (type == TYPE_OBJECT) {
			file.writeObject(data);
		} else {
			throw new IOException("Not a valid output file");
		}
	}

	public static Object readFromFile(InputFile file) throws IOException {
		Object data = null;
		if (file != null) {
			String ext = CustomFileFilter.getExtension(file.getName());
			if (ext.equalsIgnoreCase("txt")) {
				data = file.readSting();
			} else {
				data = file.readObject();
			}
		} else {
			throw new IOException("Not a valid output file");
		}
		return data;
	}

	public static int getType(InputFile infile) {
		if (infile != null) {
			return getType(infile.getName());
		} else {
			return TYPE_UNKNOWN;
		}
	}

	public static int getType(OutputFile outfile) {
		if (outfile != null) {
			return getType(outfile.getFile().getName());
		} else {
			return TYPE_UNKNOWN;
		}
	}

	private static int getType(String filename) {
		int type = TYPE_UNKNOWN;
		if (filename != null) {
			String ext = CustomFileFilter.getExtension(filename);
			if (ext.equalsIgnoreCase("txt")) {
				type = TYPE_TEXT;
			} else {
				type = TYPE_OBJECT;
			}
		}
		return type;
	}
}
