/*
 * Latex.java       Jan 3, 2007
 *
 * Copyright (c) 2007, Rod Howell, all rights reserved.
 */
package edu.ksu.cis.ndpc.io;

import edu.ksu.cis.ndpc.LangSymbols;
import edu.ksu.cis.ndpc.PredLSymbols;
import edu.ksu.cis.ndpc.PropLSymbols;
import edu.ksu.cis.ndpc.gui.RawProof;
import edu.ksu.cis.ndpc.gui.RawStatement;
import edu.ksu.cis.ndpc.rules.Rule;
import edu.ksu.cis.ndpc.Symbol;
import edu.ksu.cis.ndpc.SymbolTokenizer;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;

/**
 * Provides support for exporting a proof to LaTeX.
 * 
 * @author Rod Howell (<a href="mailto:rhowell@ksu.edu">rhowell@ksu.edu</a>)
 * 
 */
public class Latex {

	/**
	 * Writes a LaTeX version of the proof to a file chosen by the user. The
	 * file produced is suitable for inclusion into a LaTeX document. The main
	 * LaTeX document must use the packages amsmath, amssymb, array, and hhline.
	 * 
	 * @param p
	 *            The proof to be written.
	 * @return The file to which the proof was written, or null if the proof was
	 *         not written.
	 * @throws IOException
	 *             If an I/O error occurs.
	 */
	public static File export(RawProof p) throws IOException {
		File fn = chooseFile(); // Get the file name from the user
		if (fn != null) {
			PrintStream out = new PrintStream(new FileOutputStream(fn));

			// The first two lines are LaTeX comments.
			out
					.println("% This file was generated by Natural Deduction Proof Checker.");
			out
					.println("% Requires packages amsmath, amssymb, array, and hhline.");
			out.println();
			// This puts a bit more space between the rows.
			out.println("\\setlength{\\extrarowheight}{4pt}");
			out.println();
			int boxHeight = getBoxHeight(p);

			// Begin a tabular environment containing a column for each of the
			// four columns
			// of a proof (number, variable, formula, justification), plus one
			// column for each
			// vertical line forming boxes.
			out.print("\\begin{tabular}{r");
			printBarColsDefn(out, boxHeight);
			out.print("lll");
			printBarColsDefn(out, boxHeight);
			out.println("}");

			int[] boxesAtStat = p.boxesAtStat(0);
			int vertLines = boxesAtStat[0];
			// If the first line is contained in a box, we need to start with a
			// horizontal
			// line.
			if (vertLines > 0) {
				out.println("\\cline{3-" + (2 * boxHeight + 4) + "}");
			}

			// Now print each line of the proof.
			for (int i = 0; i < p.size(); i++) {

				// First, the statement number.
				out.print((i + 1) + " & ");
				for (int j = 0; j < vertLines; j++) {
					out.print("& "); // Tab over columns containing vertical
										// lines.
				}

				// White out any remaining columns reserved for vertical lines.
				String whitespace = "\\multicolumn{" + (boxHeight - vertLines)
						+ "}{c@{}}{}";
				if (boxHeight - vertLines > 0) {
					out.print(whitespace + " & ");
				}
				RawStatement s = p.getStatement(i);

				// Print the variable (if there is one) in math mode.
				String var = s.getVariable();
				if (var != null) {
					out.print("$" + var + "$");
				}
				out.print(" & ");

				// Print the formula.
				String f = s.getFormula();
				if (f != null) {
					System.out.println("Proof type = " + p.getType());
					if (p.getType() == RawProof.PROOF_PRED) {
						formatFormula(f, new PredLSymbols(), out);
					} else {
						formatFormula(f, new PropLSymbols(), out);
					}
				}
				out.print(" & ");

				// Print the rule, followed by the line numbers referenced.
				String rule = s.getRule();
				if (rule != null) {
					formatRule(rule, out);
				}
				String nums = s.getLineref();
				if (nums != null) {
					out.print(nums);
				}

				// Again, white out any columns in which vertical lines will not
				// appear,
				// then tab over the vertical lines.
				if (boxHeight - vertLines > 0) {
					out.print(" & " + whitespace);
				}
				for (int j = 0; j < vertLines; j++) {
					out.print(" &");
				}

				// End of line.
				out.println(" \\\\");

				// We now need to print any horziontal lines following this
				// statement.
				int[] boxesAtNext = p.boxesAtStat(i + 1);
				if (boxesAtStat[1] > 0 && boxesAtNext[0] > 0) {
					printDouble(boxHeight, vertLines, out); // Two lines needed.
				} else if (boxesAtStat[1] > 0) {
					// Boxes ending, but no boxes beginning.
					out.println("\\cline{" + (3 + vertLines - boxesAtStat[1])
							+ "-"
							+ (2 * boxHeight + 4 - vertLines + boxesAtStat[1])
							+ "}");
				} else if (boxesAtNext[0] > 0) {
					// Boxes beginning, but no boxes ending.
					out.println("\\cline{" + (3 + vertLines) + "-"
							+ (2 * boxHeight + 4 - vertLines) + "}");
				}
				vertLines += boxesAtNext[0] - boxesAtStat[1];
				boxesAtStat = boxesAtNext;
			}

			// End the tabular environment.
			out.println("\\end{tabular}");
			out.println();

			// Set the extra row height back to 0. This is ugly because it may
			// not
			// have been 0 initially, but I don't see a better way to do it
			// without
			// defining a package.
			out.println("\\setlength{\\extrarowheight}{0pt}");
			out.flush();
			out.close();
		}
		return fn;
	}

	/**
	 * Gets a file name from the user.
	 * 
	 * @return The file chosen, or null if no file is chosen.
	 */
	public static File chooseFile() {
		File file = null;
		int returnVal = CustomFileChooser.CANCEL_OPTION;
		CustomFileChooser fc = getFileChooser();
		returnVal = fc.showDialog(null, "Save As");
		if (returnVal == CustomFileChooser.APPROVE_OPTION) {
			file = fc.getSelectedFile();
			String fileExt = CustomFileFilter.getExtension(file);
			String selectExt = CustomFileFilter
					.getExtension((CustomFileFilter) fc.getFileFilter());
			// System.out.println("Path:"+file.getParentFile().getAbsolutePath());
			if (fileExt == null || !fileExt.equalsIgnoreCase(selectExt)) {
				file = new File(file.getParentFile().getAbsolutePath()
						+ File.separator + file.getName() + "." + selectExt);
			}
			// System.out.println("File: " + file.getAbsolutePath());
			IO.currentDirectory = file.getParent();
		}
		fc = null;
		return file;
	}

	/**
	 * Builds a CustomFileChooser to select a .tex file for exporting.
	 * 
	 * @return The CustomFileChooser.
	 */
	private static CustomFileChooser getFileChooser() {
		CustomFileChooser fc = new CustomFileChooser(IO.currentDirectory);
		fc.setDialogTitle("Export Proof");
		fc.setType(CustomFileChooser.TYPE_SAVEAS);
		fc.setAcceptAllFileFilterUsed(false);
		fc.addChoosableFileFilter(new CustomFileFilter("tex", "LaTeX Files"));
		return fc;
	}

	/**
	 * Computes the maximum depth of box nesting within the given proof.
	 * 
	 * @param p
	 *            The proof.
	 * @return The maximum box-nesting depth.
	 */
	private static int getBoxHeight(RawProof p) {
		int numStatements = p.size();
		int h = 0;
		int depth = 0;
		for (int i = 0; i < numStatements; i++) {
			int[] stats = p.boxesAtStat(i);
			depth += stats[0] - stats[1];
			if (depth > h)
				h = depth;
		}
		return h;
	}

	/**
	 * Prints a portion of a tabular environment defintion to the given stream.
	 * The output consists of a sequence of column definitions, one for each
	 * vertical line on one side of the proof.
	 * 
	 * @param out
	 *            The output stream.
	 * @param height
	 *            The number of vertical lines needed.
	 */
	private static void printBarColsDefn(PrintStream out, int height) {
		for (int i = 0; i < height; i++) {
			out.print("c@{}|");
		}
	}

	/**
	 * Prints the given formula to the given stream.
	 * 
	 * @param f
	 *            The formula.
	 * @param ls
	 *            The special symbols used.
	 * @param out
	 *            The output stream.
	 */
	private static void formatFormula(String f, LangSymbols ls, PrintStream out) {
		SymbolTokenizer tok = new SymbolTokenizer(f, ls);
		out.print("$"); // Begin math mode.
		while (tok.nextSymbol()) {
			Symbol s = tok.getSymbol();
			switch (s.getType()) {
			case LangSymbols.S_IMPL:
				out.print("\\rightarrow ");
				break;
			case LangSymbols.S_DIMPL:
				out.print("\\leftrightarrow ");
				break;
			case LangSymbols.S_DISJ:
				out.print("\\vee ");
				break;
			case LangSymbols.S_CONJ:
				out.print("\\wedge ");
				break;
			case LangSymbols.S_NEG:
				out.print("\\neg ");
				break;
			case LangSymbols.S_BOTTOM:
				out.print("\\bot ");
				break;
			case LangSymbols.S_FORALL:
				out.print("\\forall ");
				break;
			case LangSymbols.S_EXISTS:
				out.print("\\exists ");
				break;
			default: // For all other symbols, no special formatting is needed.
				out.print(s.getValue());
			}
		}
		out.print("$"); // End math mode.
	}

	/**
	 * Prints the given rule to the given stream.
	 * 
	 * @param r
	 *            The rule.
	 * @param out
	 *            The output stream.
	 */
	private static void formatRule(String r, PrintStream out) {
		switch (Rule.identify(r)) {
		case Rule.DIMP_ELIM1:
			out.print("$\\leftrightarrow$e$_1$ ");
			break;
		case Rule.DIMP_ELIM2:
			out.print("$\\leftrightarrow$e$_2$ ");
			break;
		case Rule.DIMP_INTRO:
			out.print("$\\leftrightarrow$i ");
			break;
		case Rule.IMP_INTRO:
			out.print("$\\rightarrow$i ");
			break;
		case Rule.IMP_ELIM:
			out.print("$\\rightarrow$e ");
			break;
		case Rule.OR_INTRO1:
			out.print("$\\vee$i$_1$ ");
			break;
		case Rule.OR_INTRO2:
			out.print("$\\vee$i$_2$ ");
			break;
		case Rule.DNEG_INTRO:
			out.print("$\\neg\\neg$i ");
			break;
		case Rule.DNEG_ELIM:
			out.print("$\\neg\\neg$e ");
			break;
		case Rule.AND_ELIM1:
			out.print("$\\wedge$e$_1$ ");
			break;
		case Rule.AND_ELIM2:
			out.print("$\\wedge$e$_2$ ");
			break;
		case Rule.AND_INTRO:
			out.print("$\\wedge$i ");
			break;
		case Rule.OR_ELIM:
			out.print("$\\vee$e ");
			break;
		case Rule.NEG_INTRO:
			out.print("$\\neg$i ");
			break;
		case Rule.NEG_ELIM:
			out.print("$\\neg$e ");
			break;
		case Rule.FORALL_ELIM:
			out.print("$\\forall$e ");
			break;
		case Rule.FORALL_INTRO:
			out.print("$\\forall$i ");
			break;
		case Rule.EXIST_ELIM:
			out.print("$\\exists$e ");
			break;
		case Rule.EXIST_INTRO:
			out.print("$\\exists$i ");
			break;
		case Rule.EQUAL_ELIM:
			out.print("${=}$e ");
			break;
		case Rule.EQUAL_INTRO:
			out.print("${=}$i ");
			break;
		case Rule.BOTTOM_ELIM:
			out.print("$\\bot$e ");
			break;
		default: // For all other rules, no formatting is needed.
			out.print(r + " ");
		}
	}

	/**
	 * Prints two horizontal lines to end and begin boxes between two lines of
	 * proof.
	 * 
	 * @param boxHeight
	 *            The maximum nesting depth of boxes in the proof.
	 * @param vertLines
	 *            The number of vertical lines containing the boxes being ended
	 *            and begun, including the outermost boxes being ended and
	 *            begun.
	 * @param out
	 *            The output stream.
	 */
	private static void printDouble(int boxHeight, int vertLines,
			PrintStream out) {
		// We use the \hhline command.
		// We skip the first two columns (the first vertical line is at the left
		// edge
		// of the second column.
		out.print("\\hhline{~~");

		// Print the vertical lines on the left.
		for (int i = 1; i < vertLines; i++) {
			out.print("|~");
		}

		// Print the left corners of the boxes, followed by double lines,
		// followed by
		// the right corners.
		out.print(":");
		for (int i = 0; i < 2 * (boxHeight - vertLines) + 4; i++) {
			out.print("=");
		}
		out.print(":");

		// Print the vertical lines on the right.
		for (int i = 1; i < vertLines; i++) {
			out.print("~|");
		}
		out.println("}"); // Closing brace for \hhline.
	}
}
